# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ...common._registration import register_shape_calculator
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common.data_types import (
    DictionaryType, FloatTensorType, Int64TensorType,
    SequenceType, StringTensorType,
)
from ..common import get_xgb_params


def calculate_xgboost_classifier_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=2)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType])
    N = operator.inputs[0].type.shape[0]

    xgb_node = operator.raw_operator
    params = get_xgb_params(xgb_node)
    booster = xgb_node.get_booster()
    atts = booster.attributes()
    ntrees = len(booster.get_dump(with_stats=True, dump_format = 'json'))
    objective = params["objective"]

    if objective == "binary:logistic":
        ncl = 2
    else:
        ncl = ntrees // params['n_estimators']
        if objective == "reg:logistic" and ncl == 1:
            ncl = 2
    classes = xgb_node.classes_
    if (np.issubdtype(classes.dtype, np.floating) or
            np.issubdtype(classes.dtype, np.integer)):
        operator.outputs[0].type = Int64TensorType(shape=[N])
    else:
        operator.outputs[0].type = StringTensorType(shape=[N])
    operator.outputs[1].type = operator.outputs[1].type = FloatTensorType([N, ncl])


register_shape_calculator('XGBClassifier', calculate_xgboost_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common.data_types import StringTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_tokenizer(scope, operator, container):
    op = operator.raw_operator
    if op is None:
        raise RuntimeError("op cannot be None.")
    # the SPARK version converts text to lowercase and applies "\\s" regexp to it
    # Here we'll tokenize and then normalize (to convert to lowercase)
    lowercase_output = scope.get_unique_variable_name('lowercase_tensor')
    container.add_node('StringNormalizer', operator.input_full_names[0], lowercase_output,
                       op_domain='ai.onnx',
                       name=scope.get_unique_operator_name('StringNormalizer'),
                       op_version=10,
                       case_change_action='LOWER')
    container.add_node('Tokenizer', lowercase_output, operator.output_full_names,
                       op_domain='com.microsoft',
                       name=scope.get_unique_operator_name('Tokenizer'),
                       op_version=1,
                       mark=0,
                       separators=[' ', '\t', '\r', '\n'],
                       pad_value='##ERROR##',
                       mincharnum=1)


register_converter('pyspark.ml.feature.Tokenizer', convert_sparkml_tokenizer)


def calculate_sparkml_tokenizer_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator,
                                 good_input_types=[StringTensorType])
    operator.outputs[0].type = StringTensorType()


register_shape_calculator('pyspark.ml.feature.Tokenizer', calculate_sparkml_tokenizer_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import copy

import numpy
from pyspark.ml.feature import MaxAbsScalerModel, MinMaxScalerModel, StandardScalerModel

from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_scaler(scope, operator, container):
    op = operator.raw_operator
    input_name = operator.inputs[0].full_name

    op_type = 'Scaler'
    attrs = {'name': scope.get_unique_operator_name(op_type)}
    if isinstance(op, StandardScalerModel):
        C = operator.inputs[0].type.shape[1]
        attrs['offset'] = op.mean.toArray() if op.getOrDefault("withMean") else [0.0] * C
        attrs['scale'] = [1.0 / x for x in op.std.toArray()] if op.getOrDefault("withStd") else [1.0] * C
    elif isinstance(op, MinMaxScalerModel):
        epsilon = 1.0e-8  # to avoid dividing by zero
        attrs['offset'] = [x for x in op.originalMin]
        max_min = [a - b for a, b in zip(op.originalMax, op.originalMin)]
        attrs['scale'] = [1.0 / (x + epsilon) for x in max_min]
    elif isinstance(op, MaxAbsScalerModel):
        C = operator.inputs[0].type.shape[1]
        attrs['offset'] = [0.] * C
        attrs['scale'] = [1.0 / x for x in op.maxAbs]
    else:
        raise ValueError('Unsupported Scaler: %s' % type(op))

    # ONNX does not convert arrays of float32.
    for k in attrs:
        v = attrs[k]
        if isinstance(v, numpy.ndarray) and v.dtype == numpy.float32:
            attrs[k] = v.astype(numpy.float64)
    container.add_node(op_type, input_name, operator.output_full_names, op_domain='ai.onnx.ml', **attrs)


register_converter('pyspark.ml.feature.StandardScalerModel', convert_sparkml_scaler)
register_converter('pyspark.ml.feature.MaxAbsScalerModel', convert_sparkml_scaler)
register_converter('pyspark.ml.feature.MinMaxScalerModel', convert_sparkml_scaler)


def calculate_sparkml_scaler_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType])

    input_shape = copy.deepcopy(operator.inputs[0].type.shape)
    operator.outputs[0].type = FloatTensorType(input_shape)


register_shape_calculator('pyspark.ml.feature.StandardScalerModel', calculate_sparkml_scaler_output_shapes)
register_shape_calculator('pyspark.ml.feature.MaxAbsScalerModel', calculate_sparkml_scaler_output_shapes)
register_shape_calculator('pyspark.ml.feature.MinMaxScalerModel', calculate_sparkml_scaler_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import coremltools
from uuid import uuid4
import onnx
from ...proto import onnx_proto
from ..common._topology import convert_topology
from ..common.onnx_ex import get_maximum_opset_supported
from ._parse import parse_coreml

# Import modules to invoke function registrations
from . import operator_converters
from . import shape_calculators
from .operator_converters import neural_network as nn_converters
from .shape_calculators import neural_network as nn_shape_calculators


def convert(model, name=None, initial_types=None, doc_string='', target_opset=None,
            targeted_onnx=onnx.__version__, custom_conversion_functions=None, custom_shape_calculators=None):
    '''
    This function converts the specified CoreML model into its ONNX counterpart. Some information such as the produced
    ONNX model name can be specified.

    :param model: A `CoreML model <https://apple.github.io/coremltools/coremlspecification/sections/Model.html#model>`_ or
        a CoreML MLModel object
    :param initial_types: A list providing some types for some root variables. Each element is a tuple of a variable
        name and a type defined in *data_types.py*.
    :param name: The name of the graph (type: GraphProto) in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2') used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is equivalent to the input CoreML model

    Example of initial types:
    Assume that 'A' and 'B' are two root variable names used in the CoreML
    model you want to convert. We can specify their types via:

    ::

        from onnxmltools.convert.common.data_types import FloatTensorType
        initial_type = [('A', FloatTensorType([40, 12, 1, 1])),
                        ('B', FloatTensorType([1, 32, 1, 1]))]
    '''
    if isinstance(model, coremltools.models.MLModel):
        spec = model.get_spec()
    else:
        spec = model

    if name is None:
        name = str(uuid4().hex)

    target_opset = target_opset if target_opset else get_maximum_opset_supported()
    # Parse CoreML model as our internal data structure (i.e., Topology)
    topology = parse_coreml(spec, initial_types, target_opset, custom_conversion_functions, custom_shape_calculators)

    # Parse CoreML description, author, and license. Those information will be attached to the final ONNX model.
    metadata = spec.description.metadata
    metadata_props = []
    if metadata:
        if not doc_string and metadata.shortDescription:
            doc_string = metadata.shortDescription  # If doc_string is not specified, we use description from CoreML
        if metadata.author:
            entry = onnx_proto.StringStringEntryProto()
            entry.key = 'author'
            entry.value = metadata.author
            metadata_props.append(entry)
        if metadata.license:
            entry = onnx_proto.StringStringEntryProto()
            entry.key = 'license'
            entry.value = metadata.license
            metadata_props.append(entry)

    # Convert our Topology object into ONNX. The outcome is an ONNX model.
    onnx_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)

    # Edit ONNX model's attributes related to CoreML's meta information
    if len(metadata_props) > 0:
        onnx_model.metadata_props.extend(metadata_props)

    return onnx_model



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_min
from ....common._registration import register_converter


def convert_min(scope, operator, container):
    apply_min(scope, operator.input_full_names, operator.output_full_names, container, operator.full_name)


register_converter('min', convert_min)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_converter


def convert_lrn(scope, operator, container):
    op_type = 'LRN'
    params = operator.raw_operator.lrn
    attrs = {'name': scope.get_unique_operator_name(op_type)}
    attrs['size'] = params.localSize
    attrs['alpha'] = float(params.alpha)
    attrs['beta'] = float(params.beta)
    attrs['bias'] = float(params.k)
    container.add_node(op_type, operator.input_full_names, operator.output_full_names, **attrs)


register_converter('lrn', convert_lrn)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import FloatTensorType, Int64TensorType
from ...common.utils import check_input_and_output_numbers
try:
    from libsvm.svm import C_SVC, NU_SVC
except ImportError:
    # Older version of libsvm.
    from svm import C_SVC, NU_SVC


def calculate_classifier_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=2)

    N = (operator.inputs[0].type.shape[0]
         if len(operator.inputs[0].type.shape) > 0 else None)
    if len(operator.outputs) != 2:
        raise RuntimeError("Expect only two outputs not {0}".format(len(operator.outputs)))
    svm_node =  operator.raw_operator
    if svm_node.is_probability_model():
        nc = svm_node.nr_class
    else:
        # libsvm produces n(n-1) raw scores.
        # onnxruntime aggregates the scores
        # but this behavior could be revisited.
        nc = svm_node.nr_class
        st = svm_node.param.svm_type
        if (st == C_SVC or st == NU_SVC) and nc > 2:
            nc = (nc * (nc-1)) // 2
    operator.outputs[0].type = Int64TensorType([N, 1])
    operator.outputs[1].type = FloatTensorType([N, nc])


register_shape_calculator('LibSvmSVC', calculate_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType, Int64TensorType, StringTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_permute_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N', C', H', W']

    Note that here [N', C', H', W'] means all possible permutations of [N, C, H, W]
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType, StringTensorType],
                                 good_output_types=[FloatTensorType, Int64TensorType, StringTensorType])

    input = operator.inputs[0]
    output = operator.outputs[0]

    axes = [int(i) for i in operator.raw_operator.permute.axis]
    input_shape = copy.deepcopy(input.type.shape)
    output.type.shape = [input_shape[a] for a in axes]


register_shape_calculator('permute', calculate_permute_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_upsample
from ....common._registration import register_converter


def convert_upsample(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import UpsampleLayerParams as Params
    params = operator.raw_operator.upsample

    if params.mode == Params.NN:
        mode = 'NEAREST'
    elif params.mode == Params.BILINEAR:
        mode = 'BILINEAR'
    else:
        raise ValueError('Unsupported interpolation model in up-sampling')

    width_scale = float(params.scalingFactor[1])
    height_scale = float(params.scalingFactor[0])

    apply_upsample(scope, operator.input_full_names[0], operator.output_full_names, container, operator_name=None,
                   mode=mode, scales=[1, 1, height_scale, width_scale])


register_converter('upsample', convert_upsample)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from .....proto import onnx_proto
from ....common._registration import register_converter
from .Reshape import apply_reshape


def extract_rnn_activation_info(activation):
    activation_type = activation.WhichOneof('NonlinearityType')
    alpha = None
    beta = None

    activation_map = {'linear': 'Affine',
                      'ReLU': 'Relu',
                      'leakyReLU': 'LeakyRelu',
                      'thresholdedReLU': 'ThresholdedRelu',
                      'PReLU': 'PRelu',
                      'tanh': 'Tanh',
                      'scaledTanh': 'ScaledTanh',
                      'sigmoid': 'Sigmoid',
                      'sigmoidHard': 'HardSigmoid',
                      'ELU': 'Elu',
                      'softsign': 'Softsign',
                      'softplus': 'Softplus',
                      'parametricSoftplus': 'ParametricSoftplus'}

    if activation_type not in activation_map:
        raise ValueError('Unsupported activation function: {}'.format(activation_type))

    # Notice that if we see a default vaulue (i.e., 0 for float), we may replace it with
    # the real default parameter for the specified activation function if necessary.
    if activation_type == 'leakyReLU':
        alpha = activation.leakyReLU.alpha
        if alpha == 0:
            alpha = 0.3
    elif activation_type == 'PReLU':
        raise RuntimeError('Unsupported activation function: {}'.format(activation_type))
    elif activation_type == 'ELU':
        alpha = activation.ELU.alpha
    elif activation_type == 'thresholdedReLU':
        alpha = activation.thresholdedReLU.alpha
        if alpha == 0:
            alpha = 1.0
    elif activation_type == 'scaledTanh':
        alpha = activation.scaledTanh.alpha
        beta = activation.scaledTanh.beta
    elif activation_type == 'linear':
        alpha = activation.linear.alpha
        beta = activation.linear.beta
        if alpha == 0:
            alpha = 1.0
    elif activation_type == 'sigmoidHard':
        alpha = activation.sigmoidHard.alpha
        beta = activation.sigmoidHard.beta
        if alpha == 0:
            alpha = 0.2
        if beta == 0:
            beta = 0.5
    elif activation_type == 'parametricSoftplus':
        raise RuntimeError('Unsupported activation function: {}'.format(activation_type))

    return activation_map[activation_type], alpha, beta


def convert_simple_rnn(scope, operator, container):
    # The RNN inputs are feature vector, X, and initial hidden state, h_init. Let C_in and C_out denote
    # the input feature length and the output dimension, respectively. Assume that S is the sequence length
    # (i.e., S-axis is the time axis of a sequence). In CorML, the shapes of X and h_init are [S, C_in] and
    # [1, C_out] respectively. In ONNX, the two shapes become [S, N, C_in] (X) and [D, N, C_out]
    # (h_init), where N is the batch size. To simulate CoreML RNN under ONNX, we need to introduce some extra
    # operators. Note that N=1 and D=1 always hold in ONNX if we are considering RNN from CoreML because there
    # is no batch size in CoreML and CoreML RNN is always uni-directional.
    #
    # Below we provide a visualization of our conversion for CoreML RNN.
    #
    # Symbols:
    #
    #  X: input features of CoreML RNN
    #  h_init: initial RNN state of CoreML
    #  Y: CoreML RNN's output. It can be [S, C_out] (if sequence_output is on ) or [1, C_out] (if sequence_output is off)
    #  Y_h: CoreML RNN's last hidden state
    #
    #  X': input features of ONNX RNN
    #  h_init': initial RNN state of ONNX
    #  Y': ONNX RNN's output
    #  Y_h': ONNX RNN's last hidden state
    #
    # Computational graph of CoreML RNN (sequence_output is on):
    #
    # X [S, C_in] ---> CoreML RNN ---> Y [S, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_out] -'     '---> Y_h [1, C_out]
    #
    # Computational graph we use for represent CoreML RNN into ONNX (sequence_output is on):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] -----> ONNX RNN --> Y' [S, 1, C_out] --> Reshape --> Y [S, C_out]
    #                                                        ^ |
    #                                                        | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C_out]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y_h [1, C_out]
    #
    # Computational graph of CoreML RNN (sequence_output is off):
    #
    # X [S, C_in] ---> CoreML RNN ---> Y [1, C_out]
    #                    ^     |
    #                    |     |
    # h_init [1, C_cou] -'     '---> Y_h [1, C_out] Note that in this case, Y=Y_h.
    #
    # Computational graph we use to represent CoreML RNN into ONNX (sequence_output is off):
    #
    # X [S, C_in] --> Reshape --> X' [S, 1, C_in] -----> ONNX RNN --> Y' [S, 1, C_out] Here Y' is useless.
    #                                                        ^ |
    #                                                        | |
    # h_init [1, C_out] --> Reshape -> h_init' [1, 1, C_out]-' '--> Y_h' [1, 1, C_out] ---> Reshape --> Y [1, C_out]
    #                                                                                                   |
    #                                                                                                   v
    #                                                                                                 Identity
    #                                                                                                   |
    #                                                                                                   v
    #                                                                                               Y_h [1, C_out]

    params = operator.raw_operator.simpleRecurrent
    input_size = params.inputVectorSize
    hidden_size = params.outputVectorSize

    X_name = operator.inputs[0].full_name
    X_reshape_name = scope.get_unique_variable_name('X')
    apply_reshape(scope, X_name, X_reshape_name, container, desired_shape=[-1, 1, input_size])

    rnn_op_name = scope.get_unique_operator_name('RNN')
    rnn_attrs = {'name': rnn_op_name}
    rnn_inputs = [X_reshape_name]

    # Load RNN's weight matrix and add it into RNN's input list
    rnn_w_name = scope.get_unique_variable_name(rnn_op_name + '_W')
    container.add_initializer(rnn_w_name, onnx_proto.TensorProto.FLOAT,
                              [1, hidden_size, input_size], params.weightMatrix.floatValue)
    rnn_inputs.append(rnn_w_name)

    # Load RNN's recursion matrix and add it into RNN's input list
    rnn_r_name = scope.get_unique_variable_name(rnn_op_name + '_R')
    container.add_initializer(rnn_r_name, onnx_proto.TensorProto.FLOAT,
                              [1, hidden_size, hidden_size], params.recursionMatrix.floatValue)
    rnn_inputs.append(rnn_r_name)

    if params.hasBiasVector:
        # Load RNN's bias vector and add it into RNN's input list
        rnn_b_name = scope.get_unique_variable_name(rnn_op_name + '_B')
        rnn_b_content = np.concatenate([params.biasVector.floatValue, np.zeros(hidden_size)]).flatten()
        container.add_initializer(rnn_b_name, onnx_proto.TensorProto.FLOAT, [1, 2 * hidden_size], rnn_b_content)
        rnn_inputs.append(rnn_b_name)
    else:
        # Input names are position-sensitive, so for optional but missing inputs, we need to provide an empty string.
        rnn_inputs.append('')

    # The input, sequence_lens, in ONNX is alwasy optional for this conversion, so here is always an empty string.
    rnn_inputs.append('')

    # If initial hidden state is provided, we add it into RNN's input list after adjusting its shape.
    if len(operator.inputs) == 2:
        rnn_h_init_reshape_name = scope.get_unique_variable_name(rnn_op_name + '_h_init')
        apply_reshape(scope, operator.inputs[1].full_name, rnn_h_init_reshape_name, container,
                      desired_shape=[1, 1, hidden_size])

        rnn_inputs.append(rnn_h_init_reshape_name)
        # Add a zero initializer to initial hidden state so that this variable becomes optional
        container.add_initializer(operator.inputs[1].full_name, onnx_proto.TensorProto.FLOAT,
                                  operator.inputs[1].type.shape,
                                  np.zeros(shape=operator.inputs[1].type.shape).flatten())
    else:
        # Input names are position-sensitive, so for optional but missing inputs, we need to provide an empty string.
        rnn_inputs.append('')

    # Add RNN's information of activation function
    activation, alpha, beta = extract_rnn_activation_info(params.activation)
    rnn_attrs['activations'] = [activation.encode('utf-8')]
    if alpha is not None:
        rnn_attrs['activation_alpha'] = [alpha]
    if beta is not None:
        rnn_attrs['activation_beta'] = [beta]

    # Set up other attributes
    rnn_attrs['direction'] = 'reverse' if params.reverseInput else 'forward'
    rnn_attrs['hidden_size'] = hidden_size

    # Set up version-dependent attributes
    if container.target_opset < 7:
        rnn_attrs['output_sequence'] = params.sequenceOutput
        op_version = 1
    else:
        op_version = 7

    # We use the collected information to build ONNX's RNN. ONNX RNN's outputs will be saved onto two intermediate
    # tensors and we will adjust them subsequently to mimic Keras output format.
    rnn_y_name = scope.get_unique_variable_name(rnn_op_name + '_Y')
    rnn_h_name = scope.get_unique_variable_name(rnn_op_name + '_Y_h')
    container.add_node('RNN', rnn_inputs, [rnn_y_name, rnn_h_name], op_version=op_version, **rnn_attrs)

    # Set up outputs' of RNN
    if params.sequenceOutput:
        # Connect ONNX's output and CoreML's output via a reshape operator
        apply_reshape(scope, rnn_y_name, operator.outputs[0].full_name, container, desired_shape=[-1, hidden_size])

        # Handel the second RNN output (aka last hidden state), which is optional.
        if len(operator.outputs) == 2:
            # Connect ONNX's output and CoreML's output via a reshape operator
            apply_reshape(scope, rnn_h_name, operator.outputs[1].full_name, container, desired_shape=[1, hidden_size])
    else:
        # According to CoreML, its two outputs are always identical, so we just need to compute one of them and produce
        # the other one using an identity operator. Note that the first ONNX RNN output is undefined in this case.

        # Reshape last hidden state's ONNX format to its CoreML format
        apply_reshape(scope, rnn_h_name, operator.outputs[0].full_name, container, desired_shape=[1, hidden_size])

        if len(operator.outputs) == 2:
            # Copy the first output to the second output
            container.add_node('Identity', operator.outputs[0].full_name, operator.outputs[1].full_name,
                               name=scope.get_unique_operator_name('Identity'))


register_converter('simpleRecurrent', convert_simple_rnn)



# SPDX-License-Identifier: Apache-2.0

from onnxconverter_common.data_types import FloatTensorType
from ..common._container import H2OModelContainer
from ..common._topology import Topology


def _parse_h2o(scope, model, inputs):
    '''
    :param scope: Scope object
    :param model: A h2o model data object
    :param inputs: A list of variables
    :return: A list of output variables which will be passed to next stage
    '''
    this_operator = scope.declare_local_operator("H2OTreeMojo", model)
    this_operator.inputs = inputs

    if model["params"]["classifier"]:
        label_variable = scope.declare_local_variable('label', FloatTensorType())
        probability_map_variable = scope.declare_local_variable('probabilities', FloatTensorType())
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(probability_map_variable)
    else:
        variable = scope.declare_local_variable('variable', FloatTensorType())
        this_operator.outputs.append(variable)
    return this_operator.outputs


def parse_h2o(model, initial_types=None, target_opset=None,
              custom_conversion_functions=None, custom_shape_calculators=None):
    raw_model_container = H2OModelContainer(model)
    topology = Topology(raw_model_container, default_batch_size='None',
                        initial_types=initial_types, target_opset=target_opset,
                        custom_conversion_functions=custom_conversion_functions,
                        custom_shape_calculators=custom_shape_calculators)
    scope = topology.declare_scope('__root__')

    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    for variable in inputs:
        raw_model_container.add_input(variable)

    outputs = _parse_h2o(scope, model, inputs)

    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



# SPDX-License-Identifier: Apache-2.0

import copy
from typing import List

from pyspark import SparkContext
from pyspark.ml.feature import StringIndexerModel

from ...common._registration import register_converter, register_shape_calculator
from ...common._topology import ModelComponentContainer, Operator, Scope, Variable
from ...common.data_types import Int64TensorType, StringTensorType
from ...common.utils import check_input_and_output_types


def convert_sparkml_string_indexer(scope: Scope, operator: Operator, container: ModelComponentContainer):
    op: StringIndexerModel = operator.raw_operator
    op_domain = "ai.onnx.ml"
    op_version = 2
    op_type = "LabelEncoder"

    labelsArray: List[List[str]]

    if SparkContext._active_spark_context.version.startswith("2."):
        labelsArray = [op.labels]
    else:
        labelsArray = op.labelsArray

    for i in range(0, len(labelsArray)):
        encoder_attrs = {
            "name": scope.get_unique_operator_name("StringIndexer_" + str(i)),
            "keys_strings": labelsArray[i],
            "values_int64s": list(range(0, len(labelsArray[i]))),
        }

        container.add_node(
            op_type,
            [operator.inputs[i].full_name],
            [operator.outputs[i].full_name],
            op_domain=op_domain,
            op_version=op_version,
            **encoder_attrs,
        )


register_converter("pyspark.ml.feature.StringIndexerModel", convert_sparkml_string_indexer)


def calculate_sparkml_string_indexer_output_shapes(operator: Operator):
    """
    This function just copy the input shape to the output because label encoder only alters input features' values, not
    their shape.
    """
    check_input_and_output_types(operator, good_input_types=[Int64TensorType, StringTensorType])
    input: Variable
    output: Variable
    for (input, output) in zip(operator.inputs, operator.outputs):
        input_shape = copy.deepcopy(input.type.shape)
        output.type = Int64TensorType(input_shape)


register_shape_calculator("pyspark.ml.feature.StringIndexerModel", calculate_sparkml_string_indexer_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import *
from ....common._registration import register_converter


def convert_unary(scope, operator, container):
    from coremltools.proto.NeuralNetwork_pb2 import UnaryFunctionLayerParams as Params

    params = operator.raw_operator.unary
    preprocessor_name = scope.get_unique_operator_name('Affine')
    preprocessed_variable_name = scope.get_unique_variable_name(preprocessor_name + '_output')
    apply_affine(scope, operator.input_full_names[0], preprocessed_variable_name, container,
                 operator_name=preprocessor_name, alpha=params.scale, beta=params.shift)

    if params.type == Params.RSQRT:
        sqrt_tensor_name = scope.get_unique_variable_name(operator.full_name + '_intra_tensor')
        apply_sqrt(scope, preprocessed_variable_name, sqrt_tensor_name, container)
        apply_reciprocal(scope, sqrt_tensor_name, operator.output_full_names, container)
    elif params.type == Params.POWER:
        exp_name = scope.get_unique_variable_name('Y')
        container.add_initializer(exp_name, onnx_proto.TensorProto.FLOAT, [], [params.alpha])

        apply_pow(scope, [preprocessed_variable_name, exp_name], operator.output_full_names, container,
                  operator_name=operator.full_name, broadcast=1)
    elif params.type == Params.THRESHOLD:
        apply_clip(scope, preprocessed_variable_name, operator.output_full_names, container,
                   operator_name=operator.full_name, min=params.alpha)
    elif params.type == Params.SQRT:
        apply_sqrt(scope, preprocessed_variable_name, operator.output_full_names, container,
                   operator_name=operator.full_name)
    elif params.type == Params.INVERSE:
        apply_reciprocal(scope, preprocessed_variable_name, operator.output_full_names, container,
                         operator_name=operator.full_name)
    elif params.type == Params.EXP:
        apply_exp(scope, preprocessed_variable_name, operator.output_full_names, container,
                  operator_name=operator.full_name)
    elif params.type == Params.LOG:
        apply_log(scope, preprocessed_variable_name, operator.output_full_names, container,
                  operator_name=operator.full_name)
    elif params.type == Params.ABS:
        apply_abs(scope, preprocessed_variable_name, operator.output_full_names, container,
                  operator_name=operator.full_name)
    else:
        raise ValueError('Unsupported unary function :{}'.format(params.type))


register_converter('unary', convert_unary)



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_reshape
from ....common._registration import register_converter


def convert_inner_product(scope, operator, container):
    params = operator.raw_operator.innerProduct

    # Apply pre-processing step if needed
    if len(operator.inputs[0].type.shape) == 4:
        # Input shape is [N, C, 1, 1]. Adjust input shape because Gemm in ONNX only takes 2-D input
        reshaped_tensor_name = scope.get_unique_variable_name(operator.inputs[0].full_name + '_reshaped')
        apply_reshape(scope, operator.inputs[0].full_name, reshaped_tensor_name, container,
                      desired_shape=[-1, int(params.inputChannels)])
        name_a = reshaped_tensor_name
    else:
        # Input shape is [N, C]. There is no pre-processing for applying ONNX operator.
        name_a = operator.inputs[0].full_name

    # Allocate the weights of dense layer
    name_b = scope.get_unique_variable_name(operator.full_name + '_B')
    shape_b = [params.outputChannels, params.inputChannels]
    container.add_initializer(name_b, onnx_proto.TensorProto.FLOAT, shape_b, params.weights.floatValue)

    # Allocate the bias of dense layer
    name_c = scope.get_unique_variable_name(operator.full_name + '_C')
    shape_c = [params.outputChannels]
    if params.hasBias:
        container.add_initializer(name_c, onnx_proto.TensorProto.FLOAT, shape_c, params.bias.floatValue)
    else:
        container.add_initializer(name_c, onnx_proto.TensorProto.FLOAT, shape_c, [0.] * shape_b[0])

    # Set up attributes for ONNX Gemm which is the counterpart of CoreML inner product layer in ONNX.
    attrs = {'name': operator.full_name}
    attrs['alpha'] = 1.0
    attrs['beta'] = 1.0
    attrs['transA'] = 0
    attrs['transB'] = 1

    # Get the correct version number for Gemm in ONNX
    if container.target_opset < 5:
        attrs['broadcast'] = 1
        op_version = 1
    elif container.target_opset < 7:
        attrs['broadcast'] = 1
        op_version = 6
    elif container.target_opset < 9:
        op_version = 7
    elif container.target_opset < 11:
        op_version = 9
    else:
        op_version = 11

    # Create the major ONNX operator, Gemm, to do CoreML inner product and possibly add shape adjustment
    if len(operator.inputs[0].type.shape) == 4:
        # Input shape is [N, C, 1, 1] so we expect output is also 4-D, [N, C', 1, 1].
        buffer_tensor_name = scope.get_unique_variable_name(operator.full_name + '_buffer')
        container.add_node('Gemm', [name_a, name_b, name_c], buffer_tensor_name, op_version=op_version, **attrs)
        apply_reshape(scope, buffer_tensor_name, operator.outputs[0].full_name, container,
                      desired_shape=[-1, int(params.outputChannels), 1, 1])
    else:
        # Input shape is [N, C], so we don't need to change Gemm's output shape.
        container.add_node('Gemm', [name_a, name_b, name_c], operator.outputs[0].full_name,
                           op_version=op_version, **attrs)

register_converter('innerProduct', convert_inner_product)



# SPDX-License-Identifier: Apache-2.0

import copy
from onnx import onnx_pb as onnx_proto
from ...common.data_types import FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_bucketizer(scope, operator, container):
    op = operator.raw_operator
    splits = op.getSplits()
    if splits[0] != -float("inf") or splits[-1] != float('inf'):
        raise RuntimeError("the Splits must include positive/negative infinity")

    input_shape = operator.inputs[0].type.shape
    reshape_data = scope.get_unique_variable_name('reshape_info_tensor')
    container.add_initializer(reshape_data, onnx_proto.TensorProto.INT64, [3],
                              [-1, input_shape[1], 1])
    outputs = []
    for split in splits[1:]:
        less_output = scope.get_unique_variable_name('less_output_tensor')
        initializer_name = 'initializer_' + str(split)
        container.add_initializer(initializer_name, onnx_proto.TensorProto.FLOAT, [1],
                                  [split])
        container.add_node('Less', [operator.inputs[0].full_name, initializer_name], less_output,
                           name=scope.get_unique_operator_name('Less'),
                           op_version=7)
        casted_output = scope.get_unique_variable_name('cast_output_tensor')
        container.add_node('Cast', less_output, casted_output,
                           name=scope.get_unique_operator_name('Cast'),
                           op_version=6,
                           to=1)
        redim_output = scope.get_unique_variable_name('reshape_output_tensor')
        container.add_node('Reshape', [casted_output, reshape_data], redim_output,
                           name=scope.get_unique_operator_name('Reshape'),
                           op_version=5)
        outputs.append(redim_output)
    concat_output = scope.get_unique_variable_name('concat_output_tensor')
    container.add_node('Concat', outputs, concat_output,
                       name=scope.get_unique_operator_name('Concat'),
                       op_version=1,
                       axis=2)
    argmax_output = scope.get_unique_variable_name('argmax_output_tensor')
    container.add_node('ArgMax', concat_output, argmax_output,
                       name=scope.get_unique_operator_name('ArgMax'),
                       op_version=1,
                       axis=2,
                       keepdims=0)
    container.add_node('Cast', argmax_output, operator.output_full_names,
                       name=scope.get_unique_operator_name('Cast'),
                       op_version=6,
                       to=1)

register_converter('pyspark.ml.feature.Bucketizer', convert_bucketizer)

def calculate_bucketizer_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator,
                                 good_input_types=[FloatTensorType],
                                 good_output_types=[FloatTensorType])
    input_type = copy.deepcopy(operator.inputs[0].type)
    for output in operator.outputs:
        output.type = input_type


register_shape_calculator('pyspark.ml.feature.Bucketizer', calculate_bucketizer_output_shapes)



# SPDX-License-Identifier: Apache-2.0

# To register converters for Core ML operators, import associated modules here.
from . import ArrayFeatureExtractor
from . import DictVectorizer
from . import FeatureVectorizer
from . import GLMClassifier
from . import GLMRegressor
from . import Identity
from . import Imputer
from . import Normalizer
from . import OneHotEncoder
from . import Scaler
from . import SVC
from . import SVR
from . import TensorToLabel
from . import TensorToProbabilityMap
from . import TreeEnsemble



# SPDX-License-Identifier: Apache-2.0

from .....proto import onnx_proto
from ....common._apply_operation import apply_add, apply_mul
from ....common._registration import register_converter


def deduce_broadcast_axis_and_shape(target_opset, shape):
    # This function is used to calculate the first axis aligned with the scalar and the scalar's ONNX shape for reduce-
    # like operators. Assuming input variable is always a 4-D tensor, we provide a few of examples. If scalar's shape
    # is [1, 2, 3] and input shape is [5, 2, 3, 8], the aligned axis is the [2] (indexed by 1 because indexes are 0-based)
    # in [5, 2, 3, 8], and the desired scalar shape in ONNX is [2, 3] # (the [1] in [1, 2, 3] is redundant and can cause
    # errors in ONNX's boardcasting). If the scaler's shape is [1], no matter what shape the input is, we leave the axis
    # "None" because ONNX operator may automatically handle it. After ONNX-1.2, we adopt Numpy-style broadcasting rule.

    if target_opset < 7:
        # Input shape is [N, C, H, W]
        if len(shape) == 1:
            if shape[0] == 1:
                # shape is [] for scalar, we don't specify axis because it's a scalar
                return None, []
            else:
                # shape is [C], alignment starting at C-axis (indexed by 1)
                return 1, shape
        elif len(shape) == 3:
            if shape[0] == 1:
                # shape is [1, H, W], alignment starting at H-axis (indexed by 2)
                return 2, [shape[1], shape[2]]
            else:
                # shape is [C, H, W], alignment starting at C-axis (indexed by 1)
                return 1, shape
    else:
        # Input shape is [N, C, H, W]
        if len(shape) == 1:
            if shape[0] == 1:
                # shape is [1] for scalar, we don't specify axis because it's a scalar
                return None, [1]
            else:
                # shape is [C], alignment starting at C-axis
                return None, [shape[0], 1, 1]
        elif len(shape) == 3:
            if shape[0] == 1:
                # shape is [1, H, W], alignment starting at C-axis
                return None, shape
            else:
                # shape is [C, H, W], alignment starting at C-axis
                return None, shape


def convert_scale(scope, operator, container):
    # In CoreML's ScaleLayer, the input is first scaled by their "scale" attribute and then a "bias" can be added.
    # Symbols:
    #  a: scale attribute in CoreML's ScaleLayer
    #  b: bias attribute in CoreML's ScaleLayer
    #  x: input
    #  y: output
    # The math formulation of ScaleLayer should be
    #  y = a * x + b
    # Therefore, our strategy of composing ScaleLayer is to have one multiplication followed by an addition.
    params = operator.raw_operator.scale
    op1_type = 'Mul'
    scale_axis, scale_shape = deduce_broadcast_axis_and_shape(container.target_opset, params.shapeScale)
    scale_name = scope.get_unique_variable_name(op1_type + '_B')
    container.add_initializer(scale_name, onnx_proto.TensorProto.FLOAT, scale_shape, params.scale.floatValue)

    # CoreML is at most 3-D, so we always turn broadcasting on.
    scale_broadcast = 1

    if not params.hasBias:
        # Create a element-wise multiplication and use it to scale the input. The first input is the variable we want
        # to scale while the second input is their multipliers.
        apply_mul(scope, [operator.inputs[0].full_name, scale_name], operator.output_full_names, container,
                  operator_name=operator.full_name, axis=scale_axis, broadcast=scale_broadcast)
    else:
        # Declare a temporal variable to store the scaled input
        intra_variable_name = scope.get_unique_variable_name(operator.inputs[0].full_name + '_scaled')
        # Create a element-wise multiplication and use it to scale the input and save the result to a temporal variable
        apply_mul(scope, [operator.inputs[0].full_name, scale_name], intra_variable_name, container,
                  operator_name=operator.full_name, axis=scale_axis, broadcast=scale_broadcast)

        # Prepare materials to build an Add operator for adding bias
        bias_axis, bias_shape = deduce_broadcast_axis_and_shape(container.target_opset, params.shapeBias)

        # CoreML is at most 3-D, so we always turn broadcasting on.
        bias_broadcast = 1

        bias_name = scope.get_unique_variable_name(operator.full_name + '_B')
        container.add_initializer(bias_name, onnx_proto.TensorProto.FLOAT, bias_shape, params.bias.floatValue)
        # As bias exists, we add the bias into the output of the multiplication and then use the output of addition
        # as the final output of this conversion.
        apply_add(scope, [intra_variable_name, bias_name], operator.output_full_names, container,
                  axis=bias_axis, broadcast=bias_broadcast)


register_converter('scale', convert_scale)



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import inspect
import os
import numpy
import pandas
from pyspark.ml.classification import LogisticRegression, LinearSVC
from pyspark.ml.linalg import VectorUDT, SparseVector
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlLogisticRegression(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_model_logistic_regression_binary_class(self):
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)
        #
        # truncate the features
        #
        self.spark.udf.register("truncateFeatures", lambda x: SparseVector(5, range(0,5), x.toArray()[125:130]),
                                VectorUDT())
        data = original_data.selectExpr("label", "truncateFeatures(features) as features")
        lr = LogisticRegression(maxIter=100, tol=0.0001)
        model = lr.fit(data)
        # the name of the input for Logistic Regression is 'features'
        C = model.numFeatures
        model_onnx = convert_sparkml(model, 'sparkml logistic regression', [('features', FloatTensorType([None, C]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        ]
        # known error in onnxruntime 0.3.0 case
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlLogisticRegression")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_linear_svc(self):
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "sample_libsvm_data.txt")
        original_data = self.spark.read.format("libsvm").load(input_path)
        #
        # truncate the features
        #
        self.spark.udf.register("truncateFeatures", lambda x: SparseVector(5, range(0,5), x.toArray()[125:130]),
                                VectorUDT())
        data = original_data.selectExpr("label", "truncateFeatures(features) as features")
        lsvc = LinearSVC(maxIter=10, regParam=0.01)
        model = lsvc.fit(data)
        # the name of the input for Logistic Regression is 'features'
        C = model.numFeatures
        model_onnx = convert_sparkml(model, 'Spark ML Linear SVC', [('features', FloatTensorType([None, C]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)
        # run the model
        predicted = model.transform(data)
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        expected = [ predicted.toPandas().prediction.values.astype(numpy.float32) ]
        paths = save_data_models(data_np, expected, model, model_onnx,
                                    basename="SparkmlLinearSVC")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import logging
import numpy as np
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator
from .tree_ensemble_common import (
    save_read_sparkml_model_data, sparkml_tree_dataset_to_sklearn,
    add_tree_to_attribute_pairs, get_default_tree_classifier_attribute_pairs)
from .tree_helper import rewrite_ids_and_process

logger = logging.getLogger("onnxmltools")


def convert_decision_tree_classifier(scope, operator, container):
    op = operator.raw_operator
    op_type = 'TreeEnsembleClassifier'

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs['name'] = scope.get_unique_operator_name(op_type)
    attrs["classlabels_int64s"] = list(range(0, op.numClasses))

    logger.info("[convert_decision_tree_classifier] save_read_sparkml_model_data")
    tree_df = save_read_sparkml_model_data(operator.raw_params['SparkSession'], op)
    logger.info("[convert_decision_tree_classifier] sparkml_tree_dataset_to_sklearn")
    tree = sparkml_tree_dataset_to_sklearn(tree_df, is_classifier=True)
    logger.info("[convert_decision_tree_classifier] add_tree_to_attribute_pairs")
    add_tree_to_attribute_pairs(attrs, True, tree, 0, 1., 0, leaf_weights_are_counts=True)
    logger.info("[convert_decision_tree_classifier] n_nodes=%d", len(attrs['nodes_nodeids']))

    # Some values appear in an array of one element instead of a float.

    new_attrs = rewrite_ids_and_process(attrs, logger)

    container.add_node(op_type, operator.input_full_names, [operator.outputs[0].full_name,
                       operator.outputs[1].full_name], op_domain='ai.onnx.ml', **new_attrs)


register_converter('pyspark.ml.classification.DecisionTreeClassificationModel', convert_decision_tree_classifier)


def calculate_decision_tree_classifier_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=[1, 2])
    check_input_and_output_types(operator, good_input_types=[FloatTensorType, Int64TensorType])
    if len(operator.inputs[0].type.shape) != 2:
        raise RuntimeError('Input must be a [N, C]-tensor')

    N = operator.inputs[0].type.shape[0]

    class_count = operator.raw_operator.numClasses
    operator.outputs[0].type = Int64TensorType(shape=[N])
    operator.outputs[1].type = FloatTensorType([N, class_count])


register_shape_calculator('pyspark.ml.classification.DecisionTreeClassificationModel',
                          calculate_decision_tree_classifier_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_identity(scope, operator, container):
    container.add_node('Identity', operator.input_full_names, operator.output_full_names, name=operator.full_name)


register_converter('identity', convert_identity)



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import onnx
import lightgbm
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ..common.utils import hummingbird_installed
from ._parse import parse_lightgbm, WrappedBooster

# Invoke the registration of all our converters and shape calculators
from . import operator_converters, shape_calculators  # noqa


def convert(model, name=None, initial_types=None, doc_string='', target_opset=None,
            targeted_onnx=onnx.__version__, custom_conversion_functions=None,
            custom_shape_calculators=None, without_onnx_ml=False, zipmap=True,
            split=None):
    '''
    This function produces an equivalent ONNX model of the given lightgbm model.
    The supported lightgbm modules are listed below.

    * `LGBMClassifiers <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html>`_
    * `LGBMRegressor <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html>`_
    * `Booster <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.Booster.html>`_

    :param model: A LightGBM model
    :param initial_types: a python list. Each element is a tuple of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto) in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2') used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for specifying the user customized shape calculator
    :param without_onnx_ml: whether to generate a model composed by ONNX operators only, or to allow the converter
    :param zipmap: remove operator ZipMap from the ONNX graph
    :param split: this parameter is usefull to reduce the level of discrepancies for
        big regression forest (number of trees > 100). lightgbm does all the computation
        with double whereas ONNX is using floats. Instead of having one single node
        TreeEnsembleRegressor, the converter splits it into multiple nodes TreeEnsembleRegressor,
        casts the output in double and before additioning all the outputs.
        The final graph is slower but keeps the discrepancies constant
        (it is proportional to the number of trees in a node TreeEnsembleRegressor).
        Parameter *split* is the number of trees per node. It could be possible to
        do the same with TreeEnsembleClassifier. However, the normalization of the
        probabilities significantly reduces the discrepancies.
    to use ONNX-ML operators as well.
    :return: An ONNX model (type: ModelProto) which is equivalent to the input lightgbm model
    '''
    if initial_types is None:
        raise ValueError('Initial types are required. See usage of convert(...) in '
                         'onnxmltools.convert.lightgbm.convert for details')
    if without_onnx_ml and not hummingbird_installed():
        raise RuntimeError(
            'Hummingbird is not installed. Please install hummingbird to use this feature: '
            'pip install hummingbird-ml')
    if isinstance(model, lightgbm.Booster):
        model = WrappedBooster(model)
    if name is None:
        name = str(uuid4().hex)

    target_opset = target_opset if target_opset else get_maximum_opset_supported()
    topology = parse_lightgbm(model, initial_types, target_opset, custom_conversion_functions,
                              custom_shape_calculators, zipmap=zipmap, split=split)
    topology.compile()
    onnx_ml_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)

    if without_onnx_ml:
        if zipmap:
            raise NotImplementedError(
                "Conversion with zipmap operator is not implemented with hummingbird-ml.")
        from hummingbird.ml import convert, constants
        extra_config = {}
        # extra_config[constants.ONNX_INITIAL_TYPES] = initial_types
        extra_config[constants.ONNX_OUTPUT_MODEL_NAME] = name
        extra_config[constants.ONNX_TARGET_OPSET] = target_opset
        onnx_model = convert(onnx_ml_model, "onnx", extra_config=extra_config).model
        return onnx_model

    return onnx_ml_model



# SPDX-License-Identifier: Apache-2.0

'''
Mapping and utility functions for Name to Spark ML operators
'''

from pyspark.ml import Transformer, Estimator
from pyspark.ml.feature import Binarizer
from pyspark.ml.feature import BucketedRandomProjectionLSHModel
from pyspark.ml.feature import Bucketizer
from pyspark.ml.feature import ChiSqSelectorModel
from pyspark.ml.feature import CountVectorizerModel
from pyspark.ml.feature import DCT
from pyspark.ml.feature import ElementwiseProduct
from pyspark.ml.feature import HashingTF
from pyspark.ml.feature import IDFModel
from pyspark.ml.feature import ImputerModel
from pyspark.ml.feature import IndexToString
from pyspark.ml.feature import MaxAbsScalerModel
from pyspark.ml.feature import MinHashLSHModel
from pyspark.ml.feature import MinMaxScalerModel
from pyspark.ml.feature import NGram
from pyspark.ml.feature import Normalizer
from pyspark.ml.feature import OneHotEncoderModel
from pyspark.ml.feature import PCAModel
from pyspark.ml.feature import PolynomialExpansion
from pyspark.ml.feature import QuantileDiscretizer
from pyspark.ml.feature import RegexTokenizer
from pyspark.ml.feature import StandardScalerModel
from pyspark.ml.feature import StopWordsRemover
from pyspark.ml.feature import StringIndexerModel
from pyspark.ml.feature import Tokenizer
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.feature import VectorIndexerModel
from pyspark.ml.feature import VectorSlicer
from pyspark.ml.feature import Word2VecModel

from pyspark.ml.classification import LinearSVCModel, RandomForestClassificationModel, GBTClassificationModel, \
    MultilayerPerceptronClassificationModel
from pyspark.ml.classification import LogisticRegressionModel
from pyspark.ml.classification import DecisionTreeClassificationModel
from pyspark.ml.classification import NaiveBayesModel
from pyspark.ml.classification import OneVsRestModel

from pyspark.ml.regression import AFTSurvivalRegressionModel, DecisionTreeRegressionModel, RandomForestRegressionModel
from pyspark.ml.regression import GBTRegressionModel
from pyspark.ml.regression import GeneralizedLinearRegressionModel
from pyspark.ml.regression import IsotonicRegressionModel
from pyspark.ml.regression import LinearRegressionModel

from pyspark.ml.clustering import BisectingKMeans
from pyspark.ml.clustering import KMeansModel
from pyspark.ml.clustering import GaussianMixture
from pyspark.ml.clustering import LDA


def build_sparkml_operator_name_map():
    res = {k: "pyspark.ml.feature." + k.__name__ for k in [
        Binarizer, BucketedRandomProjectionLSHModel, Bucketizer,
        ChiSqSelectorModel, CountVectorizerModel, DCT, ElementwiseProduct, HashingTF, IDFModel, ImputerModel,
        IndexToString, MaxAbsScalerModel, MinHashLSHModel, MinMaxScalerModel, NGram, Normalizer, OneHotEncoderModel,
        PCAModel, PolynomialExpansion, QuantileDiscretizer, RegexTokenizer,
        StandardScalerModel, StopWordsRemover, StringIndexerModel, Tokenizer, VectorAssembler, VectorIndexerModel,
        VectorSlicer, Word2VecModel
    ]}
    res.update({k: "pyspark.ml.classification." + k.__name__ for k in [
        LinearSVCModel, LogisticRegressionModel, DecisionTreeClassificationModel, GBTClassificationModel,
        RandomForestClassificationModel, NaiveBayesModel, MultilayerPerceptronClassificationModel, OneVsRestModel
    ]})
    res.update({k: "pyspark.ml.regression." + k.__name__ for k in [
        AFTSurvivalRegressionModel, DecisionTreeRegressionModel, GBTRegressionModel, GBTRegressionModel,
        GeneralizedLinearRegressionModel, IsotonicRegressionModel, LinearRegressionModel, RandomForestRegressionModel
    ]})
    res.update({k: "pyspark.ml.clustering." + k.__name__ for k in [
        KMeansModel
    ]})
    return res


sparkml_operator_name_map = build_sparkml_operator_name_map()


def get_sparkml_operator_name(model_type):
    '''
    Get operator name of the input argument

    :param model_type:  A spark-ml object (LinearRegression, StringIndexer, ...)
    :return: A string which stands for the type of the input model in our conversion framework
    '''
    if not issubclass(model_type, Transformer):
        if issubclass(model_type, Estimator):
            raise ValueError("Estimator must be fitted before being converted to ONNX")
        else:
            raise ValueError("Unknown model type: {}".format(model_type))
    
    if model_type not in sparkml_operator_name_map:
        raise ValueError("No proper operator name found for '%s'" % model_type)
    return sparkml_operator_name_map[model_type]




# SPDX-License-Identifier: Apache-2.0

import pickle
import os
import warnings
import numpy
import packaging.version as pv
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from ..convert.common.data_types import FloatTensorType
from .utils_backend import compare_backend, extract_options, evaluate_condition, is_backend_enabled


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


def dump_data_and_model(data, model, onnx=None, basename="model", folder=None,
                        inputs=None, backend="onnxruntime", context=None,
                        allow_failure=None, verbose=False):
    """
    Saves data with pickle, saves the model with pickle and *onnx*,
    runs and saves the predictions for the given model.
    This function is used to test a backend (runtime) for *onnx*.

    :param data: any kind of data
    :param model: any model
    :param onnx: *onnx* model or *None* to use *onnxmltools* to convert it
        only if the model accepts one float vector
    :param basemodel: three files are writen ``<basename>.data.pkl``,
        ``<basename>.model.pkl``, ``<basename>.model.onnx``
    :param folder: files are written in this folder,
        it is created if it does not exist, if *folder* is None,
        it looks first in environment variable ``ONNXTESTDUMP``,
        otherwise, it is placed into ``'tests'``.
    :param inputs: standard type or specific one if specified, only used is
        parameter *onnx* is None
    :param backend: backend used to compare expected output and runtime output.
        Two options are currently supported: None for no test,
        `'onnxruntime'` to use module *onnxruntime*.
    :param context: used if the model contains a custom operator
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :param verbose: prints more information when it fails
    :return: the created files

    Some convention for the name,
    *Bin* for a binary classifier, *Mcl* for a multiclass
    classifier, *Reg* for a regressor, *MRg* for a multi-regressor.
    The name can contain some flags. Expected outputs refer to the
    outputs computed with the original library, computed outputs
    refer to the outputs computed with a ONNX runtime.

    * ``-CannotLoad``: the model can be converted but the runtime cannot load it
    * ``-Dec3``: compares expected and computed outputs up to 3 decimals (5 by default)
    * ``-Dec4``: compares expected and computed outputs up to 4 decimals (5 by default)
    * ``-NoProb``: The original models computed probabilites for two classes *size=(N, 2)*
      but the runtime produces a vector of size *N*, the test will compare the second column
      to the column
    * ``-OneOff``: the ONNX runtime cannot computed the prediction for several inputs,
      it must be called for each of them
      and computed output.
    * ``-Out0``: only compares the first output on both sides
    * ``-Reshape``: merges all outputs into one single vector and resizes it before comparing
    * ``-SkipDim1``: before comparing expected and computed output,
      arrays with a shape like *(2, 1, 2)* becomes *(2, 2)*

    If the *backend* is not None, the function either raises an exception
    if the comparison between the expected outputs and the backend outputs
    fails or it saves the backend output and adds it to the results.
    """
    runtime_test = dict(model=model, data=data)

    if folder is None:
        folder = os.environ.get('ONNXTESTDUMP', 'tests/temp')
    if not os.path.exists(folder):
        os.makedirs(folder)

    if hasattr(model, "predict"):
        import lightgbm
        import xgboost
        if isinstance(model, lightgbm.Booster):
            # LightGBM Booster
            model_dict = model.dump_model()
            if model_dict['objective'].startswith('binary'):
                score = model.predict(data)
                if len(score.shape) < 2 or score.shape[1] == 1:
                    score = score.ravel()
                    score = numpy.vstack([1-score, score]).T
                prediction = [score[:, 1] > 0.5, score]
            elif model_dict['objective'].startswith('multiclass'):
                score = model.predict(data)
                prediction = [score.argmax(axis=1), score]
            else:
                prediction = [model.predict(data)]
        elif isinstance(model, xgboost.Booster):
            # XGBoost Booster
            from ..convert.xgboost._parse import _get_attributes
            from xgboost import DMatrix
            datax = DMatrix(data)
            model_dict = _get_attributes(model)
            if model_dict['objective'].startswith('binary'):
                score = model.predict(datax)
                prediction = [score > 0.5, numpy.vstack([1-score, score]).T]
            elif model_dict['objective'].startswith('multi:softprob'):
                score = model.predict(datax)
                prediction = [score.argmax(axis=1), score]
            elif model_dict['objective'].startswith('multi:softmax'):
                score = model.predict(datax, output_margin=True)
                prediction = [score.argmax(axis=1), score]
            else:
                prediction = [model.predict(datax)]
        elif hasattr(model, "predict_proba"):
            # Classifier
            if hasattr(model, 'get_params'):
                params = model.get_params()
                if 'objective' in params:
                    objective = params['objective']
                    if objective == "multi:softmax":
                        prediction = [model.predict(data)]
                    else:
                        prediction = [model.predict(data), model.predict_proba(data)]
                else:
                    prediction = [model.predict(data), model.predict_proba(data)]
            else:
                prediction = [model.predict(data), model.predict_proba(data)]
        elif hasattr(model, "predict_with_probabilities"):
            # Classifier that returns all in one go
            prediction = model.predict_with_probabilities(data)
        elif hasattr(model, "decision_function"):
            # Classifier without probabilities
            prediction = [model.predict(data), model.decision_function(data)]
        elif hasattr(model, "layers"):
            # Keras
            if len(model.input_names) != 1:
                raise NotImplemented("Only neural network with one input are supported")
            prediction = [model.predict(data)]
        else:
            # Regressor
            prediction = [model.predict(data)]
    elif hasattr(model, "transform"):
        prediction = model.transform(data)
    else:
        raise TypeError("Model has not predict or transform method: {0}".format(type(model)))

    runtime_test['expected'] = prediction

    names = []
    dest = os.path.join(folder, basename + ".expected.pkl")
    names.append(dest)
    with open(dest, "wb") as f:
        pickle.dump(prediction, f)

    dest = os.path.join(folder, basename + ".data.pkl")
    names.append(dest)
    with open(dest, "wb") as f:
        pickle.dump(data, f)

    dest = os.path.join(folder, basename + ".model.pkl")
    names.append(dest)
    with open(dest, "wb") as f:
        pickle.dump(model, f)

    if onnx is None:
        array = numpy.array(data)
        if inputs is None:
            inputs = [('input', FloatTensorType(list(array.shape)))]
        onnx, _ = convert_model(model, basename, inputs)

    dest = os.path.join(folder, basename + ".model.onnx")
    names.append(dest)
    with open(dest, "wb") as f:
        f.write(onnx.SerializeToString())

    runtime_test["onnx"] = dest

    # backend
    if backend is not None:
        if not isinstance(backend, list):
            backend = [backend]
        for b in backend:
            if not is_backend_enabled(b):
                continue
            if isinstance(allow_failure, str):
                allow = evaluate_condition(b, allow_failure)
            else:
                allow = allow_failure
            if allow is None:
                output = compare_backend(b, runtime_test, options=extract_options(basename),
                                         context=context, verbose=verbose)
            else:
                try:
                    output = compare_backend(b, runtime_test, options=extract_options(basename),
                                             context=context, verbose=verbose)
                except AssertionError as e:
                    if isinstance(allow, bool) and allow:
                        warnings.warn("Issue with '{0}' due to {1}".format(basename, e))
                        continue
                    else:
                        raise e
            if output is not None:
                dest = os.path.join(folder, basename + ".backend.{0}.pkl".format(b))
                names.append(dest)
                with open(dest, "wb") as f:
                    pickle.dump(output, f)

    return names


def convert_model(model, name, input_types, without_onnx_ml=False, **kwargs):
    """
    Runs the appropriate conversion method.

    :param model: model
    :return: *onnx* model
    """
    from sklearn.base import BaseEstimator
    if model.__class__.__name__.startswith("LGBM"):
        from onnxmltools.convert import convert_lightgbm
        model, prefix = convert_lightgbm(model, name, input_types, without_onnx_ml=without_onnx_ml, **kwargs), "LightGbm"
    elif model.__class__.__name__.startswith("XGB"):
        from onnxmltools.convert import convert_xgboost
        model, prefix = convert_xgboost(model, name, input_types, **kwargs), "XGB"
    elif model.__class__.__name__ == 'Booster':
        import lightgbm
        if isinstance(model, lightgbm.Booster):
            from onnxmltools.convert import convert_lightgbm
            model, prefix = convert_lightgbm(model, name, input_types, without_onnx_ml=without_onnx_ml, **kwargs), "LightGbm"
        else:
            raise RuntimeError("Unable to convert model of type '{0}'.".format(type(model)))
    elif model.__class__.__name__.startswith("CatBoost"):
        from onnxmltools.convert import convert_catboost
        model, prefix = convert_catboost(model, name, input_types, **kwargs), "CatBoost"
    elif isinstance(model, BaseEstimator):
        from onnxmltools.convert import convert_sklearn
        model, prefix = convert_sklearn(model, name, input_types, **kwargs), "Sklearn"
    else:
        from onnxmltools.convert import convert_coreml
        model, prefix = convert_coreml(model, name, input_types, **kwargs), "Cml"
    if model is None:
        raise RuntimeError("Unable to convert model of type '{0}'.".format(type(model)))
    return model, prefix


def dump_one_class_classification(model, suffix="", folder=None, allow_failure=None):
    """
    Trains and dumps a model for a One Class outlier problem.
    The function trains a model and calls
    :func:`dump_data_and_model`.

    :param model: any model following *scikit-learn* API
    :param suffix: added to filenames
    :param folder: where to save the file
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :return: output of :func:`dump_data_and_model`

    Every created filename will follow the pattern:
    ``<folder>/<prefix><task><classifier-name><suffix>.<data|expected|model|onnx>.<pkl|onnx>``.
    """
    X = [[0., 1.], [1., 1.], [2., 0.]]
    X = numpy.array(X, dtype=numpy.float32)
    y = [1, 1, 1]
    model.fit(X, y)
    model_onnx, prefix = convert_model(model, 'one_class', [('input', FloatTensorType([None, 2]))],
                                       target_opset=TARGET_OPSET)
    return dump_data_and_model(X, model, model_onnx, folder=folder, allow_failure=allow_failure,
                               basename=prefix + "One" + model.__class__.__name__ + suffix)


def dump_binary_classification(model, suffix="", folder=None, allow_failure=None, verbose=False):
    """
    Trains and dumps a model for a binary classification problem.

    :param model: any model following *scikit-learn* API
    :param suffix: added to filenames
    :param folder: where to save the file
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :param verbose: prints more information when it fails
    :return: output of :func:`dump_data_and_model`

    Every created filename will follow the pattern:
    ``<folder>/<prefix><task><classifier-name><suffix>.<data|expected|model|onnx>.<pkl|onnx>``.
    """
    X = [[0, 1], [1, 1], [2, 0]]
    X = numpy.array(X, dtype=numpy.float32)
    y = [0, 1, 0]
    model.fit(X, y)
    model_onnx, prefix = convert_model(model, 'tree-based binary classifier', [('input', FloatTensorType([None, 2]))],
                                       target_opset=TARGET_OPSET)
    dump_data_and_model(X, model, model_onnx, folder=folder, allow_failure=allow_failure,
                        basename=prefix + "Bin" + model.__class__.__name__ + suffix,
                        verbose=verbose)

def dump_multiple_classification(model, suffix="", folder=None, allow_failure=None):
    """
    Trains and dumps a model for a binary classification problem.

    :param model: any model following *scikit-learn* API
    :param suffix: added to filenames
    :param folder: where to save the file
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :return: output of :func:`dump_data_and_model`

    Every created filename will follow the pattern:
    ``<folder>/<prefix><task><classifier-name><suffix>.<data|expected|model|onnx>.<pkl|onnx>``.
    """
    X = [[0, 1], [1, 1], [2, 0], [0.5, 0.5], [1.1, 1.1], [2.1, 0.1]]
    X = numpy.array(X, dtype=numpy.float32)
    y = [0, 1, 2, 1, 1, 2]
    model.fit(X, y)
    model_onnx, prefix = convert_model(model, 'tree-based multi-output regressor', [('input', FloatTensorType([None, 2]))],
                                       target_opset=TARGET_OPSET)
    dump_data_and_model(X, model, model_onnx, folder=folder, allow_failure=allow_failure,
                        basename=prefix + "Mcl" + model.__class__.__name__ + suffix)


def dump_multiple_regression(model, suffix="", folder=None, allow_failure=None):
    """
    Trains and dumps a model for a multi regression problem.

    :param model: any model following *scikit-learn* API
    :param suffix: added to filenames
    :param folder: where to save the file
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :return: output of :func:`dump_data_and_model`

    Every created filename will follow the pattern:
    ``<folder>/<prefix><task><classifier-name><suffix>.<data|expected|model|onnx>.<pkl|onnx>``.
    """
    X = [[0, 1], [1, 1], [2, 0]]
    X = numpy.array(X, dtype=numpy.float32)
    y = numpy.array([[100, 50], [100, 49], [100, 99]], dtype=numpy.float32)
    model.fit(X, y)
    model_onnx, prefix = convert_model(model, 'tree-based multi-output regressor', [('input', FloatTensorType([None, 2]))],
                                       target_opset=TARGET_OPSET)
    dump_data_and_model(X, model, model_onnx, folder=folder, allow_failure=allow_failure,
                        basename=prefix + "MRg" + model.__class__.__name__ + suffix)


def dump_single_regression(model, suffix="", folder=None, allow_failure=None):
    """
    Trains and dumps a model for a regression problem.

    :param model: any model following *scikit-learn* API
    :param prefix: library name
    :param suffix: added to filenames
    :param folder: where to save the file
    :param allow_failure: None to raise an exception if comparison fails
        for the backends, otherwise a string which is then evaluated to check
        whether or not the test can fail, example:
        ``"pv.Version(onnx.__version__) < pv.Version('1.3.0')"``
    :return: output of :func:`dump_data_and_model`

    Every created filename will follow the pattern:
    ``<folder>/<prefix><task><classifier-name><suffix>.<data|expected|model|onnx>.<pkl|onnx>``.
    """
    X = [[0, 1], [1, 1], [2, 0]]
    X = numpy.array(X, dtype=numpy.float32)
    y = numpy.array([100, -10, 50], dtype=numpy.float32)
    model.fit(X, y)
    model_onnx, prefix = convert_model(model, 'tree-based regressor', [('input', FloatTensorType([None, 2]))],
                                       target_opset=TARGET_OPSET)
    dump_data_and_model(X, model, model_onnx, folder=folder, allow_failure=allow_failure,
                        basename=prefix + "Reg" + model.__class__.__name__ + suffix)


def make_report_backend(folder):
    """
    Looks into a folder for dumped files after
    the unit tests.
    """
    res = {}
    files = os.listdir(folder)
    for name in files:
        if name.endswith(".expected.pkl"):
            model = name.split(".")[0]
            if model not in res:
                res[model] = {}
            res[model]["_tested"] = True
        elif '.backend.' in name:
            bk = name.split(".backend.")[-1].split(".")[0]
            model = name.split(".")[0]
            if model not in res:
                res[model] = {}
            res[model][bk] = True

    def dict_update(d, u):
        d.update(u)
        return d

    aslist = [dict_update(dict(_model=k), v) for k, v in res.items()]
    return aslist
